@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # Copyright 2014 Objectif Libre
 #
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
@@ -13,9 +12,9 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 #
-# @author: Stéphane Albert
-#
+from datetime import timedelta
 import decimal
+import functools
 import hashlib
 import multiprocessing
 import random
@@ -23,7 +22,8 @@ import sys
 import time
 
 import cotyledon
-import eventlet
+import futurist
+from futurist import waiters
 from oslo_concurrency import lockutils
 from oslo_config import cfg
 from oslo_log import log as logging
@@ -34,12 +34,13 @@ from tooz import coordination
 
 from cloudkitty import collector
 from cloudkitty import config  # noqa
+from cloudkitty import dataframe
 from cloudkitty import extension_manager
 from cloudkitty import messaging
 from cloudkitty import storage
 from cloudkitty import storage_state as state
-from cloudkitty import transformer
 from cloudkitty import utils as ck_utils
+from cloudkitty.utils import tz as tzutils
 
 
 LOG = logging.getLogger(__name__)
@@ -55,10 +56,18 @@ orchestrator_opts = [
     cfg.IntOpt(
         'max_workers',
         default=multiprocessing.cpu_count(),
+        sample_default=4,
         min=1,
         help='Max nb of workers to run. Defaults to the nb of available CPUs'),
-    cfg.IntOpt('max_greenthreads', default=100, min=1,
-               help='Maximal number of greenthreads to use per worker.'),
+    cfg.IntOpt('max_threads',
+               # NOTE(peschk_l): This is the futurist default
+               default=multiprocessing.cpu_count() * 5,
+               sample_default=20,
+               min=1,
+               deprecated_name='max_greenthreads',
+               advanced=True,
+               help='Maximal number of threads to use per worker. Defaults to '
+               '5 times the nb of available CPUs'),
 ]
 
 CONF.register_opts(orchestrator_opts, group='orchestrator')
@@ -106,9 +115,41 @@ class RatingEndpoint(object):
             return module_list
 
     def quote(self, ctxt, res_data):
-        LOG.debug('Received quote from RPC.')
+        LOG.debug('Received quote request [%s] from RPC.', res_data)
         worker = APIWorker()
-        return str(worker.quote(res_data))
+
+        start = tzutils.localized_now()
+        end = tzutils.add_delta(start, timedelta(seconds=CONF.collect.period))
+
+        # Need to prepare data to support the V2 processing format
+        usage = {}
+        for k in res_data['usage']:
+            all_data_points_for_metric = []
+            all_quote_data_entries = res_data['usage'][k]
+            for p in all_quote_data_entries:
+                vol = p['vol']
+                desc = p.get('desc', {})
+
+                data_point = dataframe.DataPoint(
+                    vol['unit'],
+                    vol['qty'],
+                    0,
+                    desc.get('groupby', []),
+                    desc.get('metadata', []),
+                )
+                all_data_points_for_metric.append(data_point)
+            usage[k] = all_data_points_for_metric
+
+        frame = dataframe.DataFrame(
+            start=start,
+            end=end,
+            usage=usage,
+        )
+
+        quote_result = worker.quote(frame)
+        LOG.debug("Quote result [%s] for input data [%s].",
+                  quote_result, res_data)
+        return str(quote_result)
 
     def reload_modules(self, ctxt):
         LOG.info('Received reload modules command.')
@@ -165,12 +206,10 @@ class ScopeEndpoint(object):
                         lock_name,
                     )
                 )
-                state_dt = ck_utils.iso2dt(res_data['state'])
+                state_dt = tzutils.dt_from_iso(res_data['state'])
                 try:
                     self._storage.delete(begin=state_dt, end=None, filters={
                         scope['scope_key']: scope['scope_id'],
-                        'collector': scope['collector'],
-                        'fetcher': scope['fetcher'],
                     })
                     self._state.set_state(
                         scope['scope_id'],
@@ -212,18 +251,23 @@ class APIWorker(BaseWorker):
         super(APIWorker, self).__init__(tenant_id)
 
     def quote(self, res_data):
+        quote_result = res_data
         for processor in self._processors:
-            processor.obj.quote(res_data)
+            quote_result = processor.obj.quote(quote_result)
 
         price = decimal.Decimal(0)
-        for res in res_data:
-            for res_usage in res['usage'].values():
-                for data in res_usage:
-                    price += data.get('rating', {}).get('price',
-                                                        decimal.Decimal(0))
+        for _, point in quote_result.iterpoints():
+            price += point.price
         return price
 
 
+def _check_state(obj, period, tenant_id):
+    timestamp = obj._state.get_state(tenant_id)
+    return ck_utils.check_time_state(timestamp,
+                                     period,
+                                     CONF.collect.wait_periods)
+
+
 class Worker(BaseWorker):
     def __init__(self, collector, storage, tenant_id, worker_id):
         self._collector = collector
@@ -232,26 +276,29 @@ class Worker(BaseWorker):
         self._wait_time = CONF.collect.wait_periods * self._period
         self._tenant_id = tenant_id
         self._worker_id = worker_id
+        self._log_prefix = '[scope: {scope}, worker: {worker}] '.format(
+            scope=self._tenant_id, worker=self._worker_id)
         self._conf = ck_utils.load_conf(CONF.collect.metrics_conf)
         self._state = state.StateManager()
+        self._check_state = functools.partial(
+            _check_state, self, self._period, self._tenant_id)
 
         super(Worker, self).__init__(self._tenant_id)
 
     def _collect(self, metric, start_timestamp):
-        next_timestamp = start_timestamp + self._period
+        next_timestamp = tzutils.add_delta(
+            start_timestamp, timedelta(seconds=self._period))
 
-        raw_data = self._collector.retrieve(
+        name, data = self._collector.retrieve(
             metric,
             start_timestamp,
             next_timestamp,
             self._tenant_id,
         )
-        if not raw_data:
+        if not data:
             raise collector.NoDataCollected
 
-        return {'period': {'begin': start_timestamp,
-                           'end': next_timestamp},
-                'usage': raw_data}
+        return name, data
 
     def _do_collection(self, metrics, timestamp):
 
@@ -260,58 +307,57 @@ class Worker(BaseWorker):
                 return self._collect(metric, timestamp)
             except collector.NoDataCollected:
                 LOG.info(
-                    '[scope: {scope}, worker: {worker}] No data collected '
+                    self._log_prefix + 'No data collected '
                     'for metric {metric} at timestamp {ts}'.format(
-                        scope=self._tenant_id,
-                        worker=self._worker_id,
-                        metric=metric,
-                        ts=ck_utils.ts2dt(timestamp))
-                )
-                return None
+                        metric=metric, ts=timestamp))
+                return metric, None
             except Exception as e:
-                LOG.warning(
-                    '[scope: {scope}, worker: {worker}] Error while collecting'
+                LOG.exception(
+                    self._log_prefix + 'Error while collecting'
                     ' metric {metric} at timestamp {ts}: {e}. Exiting.'.format(
-                        scope=self._tenant_id,
-                        worker=self._worker_id,
-                        metric=metric,
-                        ts=ck_utils.ts2dt(timestamp),
-                        e=e)
-                )
+                        metric=metric, ts=timestamp, e=e))
                 # FIXME(peschk_l): here we just exit, and the
                 # collection will be retried during the next collect
                 # cycle. In the future, we should implement a retrying
                 # system in workers
                 sys.exit(1)
 
-        return list(filter(
-            lambda x: x is not None,
-            eventlet.GreenPool(size=CONF.orchestrator.max_greenthreads).imap(
-                _get_result, metrics)))
-
-    def check_state(self):
-        timestamp = self._state.get_state(self._tenant_id)
-        return ck_utils.check_time_state(timestamp,
-                                         self._period,
-                                         CONF.collect.wait_periods)
+        with futurist.ThreadPoolExecutor(
+                max_workers=CONF.orchestrator.max_threads) as tpool:
+            futs = [tpool.submit(_get_result, metric) for metric in metrics]
+            LOG.debug(self._log_prefix +
+                      'Collecting {} metrics.'.format(len(metrics)))
+            results = [r.result() for r in waiters.wait_for_all(futs).done]
+            LOG.debug(self._log_prefix + 'Collecting {} metrics took {}s '
+                      'total, with {}s average'.format(
+                          tpool.statistics.executed,
+                          tpool.statistics.runtime,
+                          tpool.statistics.average_runtime))
+        return dict(filter(lambda x: x[1] is not None, results))
 
     def run(self):
         while True:
-            timestamp = self.check_state()
+            timestamp = self._check_state()
             if not timestamp:
                 break
 
             metrics = list(self._conf['metrics'].keys())
 
             # Collection
-            data = self._do_collection(metrics, timestamp)
+            usage_data = self._do_collection(metrics, timestamp)
 
+            frame = dataframe.DataFrame(
+                start=timestamp,
+                end=tzutils.add_delta(timestamp,
+                                      timedelta(seconds=self._period)),
+                usage=usage_data,
+            )
             # Rating
             for processor in self._processors:
-                processor.obj.process(data)
+                frame = processor.obj.process(frame)
 
             # Writing
-            self._storage.push(data, self._tenant_id)
+            self._storage.push([frame], self._tenant_id)
             self._state.set_state(self._tenant_id, timestamp)
 
 
@@ -326,8 +372,7 @@ class Orchestrator(cotyledon.Service):
             invoke_on_load=True,
         ).driver
 
-        transformers = transformer.get_transformers()
-        self.collector = collector.get_collector(transformers)
+        self.collector = collector.get_collector()
         self.storage = storage.get_storage()
         self._state = state.StateManager()
 
@@ -342,6 +387,8 @@ class Orchestrator(cotyledon.Service):
             CONF.orchestrator.coordination_url,
             uuidutils.generate_uuid().encode('ascii'))
         self.coord.start(start_heart=True)
+        self._check_state = functools.partial(
+            _check_state, self, CONF.collect.period)
 
     def _init_messaging(self):
         target = oslo_messaging.Target(topic='cloudkitty',
@@ -354,12 +401,6 @@ class Orchestrator(cotyledon.Service):
         self.server = messaging.get_server(target, endpoints)
         self.server.start()
 
-    def _check_state(self, tenant_id):
-        timestamp = self._state.get_state(tenant_id)
-        return ck_utils.check_time_state(timestamp,
-                                         CONF.collect.period,
-                                         CONF.collect.wait_periods)
-
     def process_messages(self):
         # TODO(sheeprine): Code kept to handle threading and asynchronous
         # reloading
@@ -372,20 +413,20 @@ class Orchestrator(cotyledon.Service):
         while True:
             self.tenants = self.fetcher.get_tenants()
             random.shuffle(self.tenants)
-            LOG.info('[Worker: {w}] Tenants loaded for fetcher {f}'.format(
-                w=self._worker_id, f=self.fetcher.name))
+            LOG.info('[Worker: {w}] {n} tenants loaded for fetcher {f}'.format(
+                w=self._worker_id, n=len(self.tenants), f=self.fetcher.name))
 
             for tenant_id in self.tenants:
 
                 lock_name, lock = get_lock(self.coord, tenant_id)
                 LOG.debug(
-                    '[Worker: {w}] Trying to acquire lock "{l}" ...'.format(
-                        w=self._worker_id, l=lock_name)
+                    '[Worker: {w}] Trying to acquire lock "{lck}" ...'.format(
+                        w=self._worker_id, lck=lock_name)
                 )
                 if lock.acquire(blocking=False):
                     LOG.debug(
-                        '[Worker: {w}] Acquired lock "{l}" ...'.format(
-                            w=self._worker_id, l=lock_name)
+                        '[Worker: {w}] Acquired lock "{lck}" ...'.format(
+                            w=self._worker_id, lck=lock_name)
                     )
                     state = self._check_state(tenant_id)
                     if state:
